"""Contains probability concept classes."""

import random
from fractions import Fraction

class EventSpace:
    """An abstract class for objects which are a container of simple events.

    Any class inheriting from this class should ensure its instances have a
    ``_simple_events`` property which is a set of :py:class:`.Simple Event`
    objects."""

    def __contains__(self, member):
        if isinstance(member, Event):
            return member.simple_events.issubset(self._simple_events)
        for event in self._simple_events:
            if event.outcome == member: return True


    @property
    def simple_events(self):
        """The set of simple events in this space.

        :rtype: ``set``"""

        return set(self._simple_events)


    def outcomes(self, p=False):
        """The set of outcomes that the event space's simple events can
        produce.

        :param bool p: if ``True``, the results will be returned as a dict with\
        probabilities associated.

        :rtype: ``set`` or ``dict``"""

        if p: return {e.outcome: e.probability() for e in self._simple_events}
        return set([e.outcome for e in self._simple_events])



class Event(EventSpace):
    """Base class: :py:class:`.EventSpace`

    An occurance that is made up of multiple simple events.

    Events are containers both of their simple events, and the outcomes of those
    simple events.

    :param \*events: The :py:class:`.Event` objects within this set.
    :param str name: The name of the event (default is 'E').
    :raises TypeError: if non-events are given.
    :raises TypeError: if the name is not a string."""

    def __init__(self, *events, name="E"):
        if any(not isinstance(e, Event) for e in events):
            raise TypeError(f"{events} contains non-events")
        if not isinstance(name, str):
            raise TypeError(f"Name {name} is not str")
        self._simple_events = set()
        for event in events: self._simple_events.update(event._simple_events)
        self._name = name


    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._name}>"


    def __or__(self, event):
        if not isinstance(event, Event):
            raise TypeError(f"{event} is not an Event")
        return Event(*(self._simple_events | event._simple_events))


    def __and__(self, event):
        if not isinstance(event, Event):
            raise TypeError(f"{event} is not an Event")
        return Event(*(self._simple_events & event._simple_events))


    @property
    def sample_space(self):
        """The sample space that the event is part of.

        :rtype: ``SampleSpace``"""

        for event in self._simple_events:
            return event._sample_space


    @property
    def name(self):
        """Returns the name of the Event.

        :rtype: ``str``"""

        return self._name


    @property
    def complement(self):
        """Returns the complement of the event - the event that this event does
        not happen.

        :rtype: ``Event``"""

        return Event(
         *(self.sample_space.simple_events - self._simple_events)
        )


    def probability(self, given=None, fraction=False):
        """Returns the probability of the event happening.

        :param Event given: an optional pre-condition to consider.
        :param bool fraction: If ``True``, the result will be returned as a\
        ``Fraction``.
        :raises TypeError: if the given event is not an :py:class:`.Event`
        :rtype: ``float``"""

        if given:
            if not isinstance(given, Event):
                raise TypeError(f"{given} is not an event")
            p = (self & given).probability(fraction=True)
            p /= given.probability(fraction=True)
        else:
            p = sum(event._probability for event in self._simple_events)
        return p if fraction else p.numerator / p.denominator


    def mutually_exclusive_with(self, event):
        """Looks at some other event and checks if this event is mutually
        exclusive with the other event. That is, whether it is impossible for
        them both to happen in a given statistical experiment.

        :param Event event: the other event to check with.
        :raises TypeError: if a non-Event is given.
        :rtype: ``bool``"""

        if not isinstance(event, Event):
            raise TypeError(f"{event} is not an event")
        return not self._simple_events & event._simple_events


    def independent_of(self, event):
        """Checks to see if this event is independent of some other event - that
        is, whether its probability is unaffacted by the occurence of the other
        event.

        :param Event event: the other event to check with.
        :raises TypeError: if a non-Event is given.
        :rtype: ``bool``"""

        if not isinstance(event, Event):
            raise TypeError(f"{event} is not an event")
        return self.probability(fraction=True) == self.probability(
         fraction=True, given=event
        )


    def dependent_on(self, event):
        """Checks to see if this event is dependent of some other event - that
        is, whether its probability is affacted by the occurence of the other
        event.

        :param Event event: the other event to check with.
        :raises TypeError: if a non-Event is given.
        :rtype: ``bool``"""

        return not self.independent_of(event)



class SimpleEvent(Event):
    """Base class: py:class:`.Event`

    A simple event - a single outcome of a statistical experiment.

    :param outcome: The result of this event occuring.
    :param float probability: The probability of this event occuring.
    :raises TypeError: if probability isn't numeric.
    :raises ValueError: if probability is not between 0 and 1."""

    def __init__(self, outcome, probability, space):
        self._outcome = self._name = outcome
        if not isinstance(probability, (int, float, Fraction)):
            raise TypeError("probability {} is not numeric".format(probability))
        if not 0 <= probability <= 1:
            raise ValueError("probability {} is invalid".format(probability))
        self._probability = probability
        self._sample_space = space
        self._simple_events = set((self,))


    @property
    def outcome(self):
        """The result of this event occuring."""

        return self._outcome



class SampleSpace(EventSpace):
    """Base class: :py:class:`.EventSpace`

    The set of all possible things that can result from a statistical
    experiment.

    :param \*outcomes: All the possible outcomes.
    :param dict p: The probabilities for the supplied outcomes. If not given,\
    these will be weighted equally.
    :raises ValueError: if you supply probabilities that don't add up to 1."""

    def __init__(self, *outcomes, p=None):
        if not outcomes and not p:
            raise ValueError("Sample spaces need at least one outcome")
        if p is None:
            p_per_event = Fraction(1, len(outcomes))
            fraction_p = {event: p_per_event for event in outcomes}
        else:
            fraction_p = {k: Fraction(str(p[k])) for k in p}
            unaccounted_outcomes = [o for o in outcomes if o not in p]
            if unaccounted_outcomes:
                remaining_p = 1 - sum(fraction_p.values())
                p_per_event = (remaining_p / len(unaccounted_outcomes))
                for e in unaccounted_outcomes:
                    fraction_p[e] = p_per_event
        if sum(fraction_p.values()) != 1:
            raise ValueError(f"Probabilities do not add up to 1: {fraction_p}")
        self._simple_events = set([
         SimpleEvent(e, fraction_p[e], self) for e in fraction_p
        ])


    def __repr__(self):
        return f"<SampleSpace ({len(self._simple_events)} simple events)>"


    def event(self, *outcomes, name=None):
        """If a single outcome is given, this function will return the
        :py:class:`.SimpleEvent` corresponding to that outcome.

        If multiple outcomes are given, the function will return the
        :py:class:`.Event` corresponding to one of those outcomes occuring.

        If a callable is provided, the function will return the
        :py:class:`.Event` of all the simple events that return ``True`` when
        the callable is applied to their outcome.

        :param \*outcomes: The outcome(s) to look for.
        :param str name: The name to apply if a :py:class:`.Event` is returned.
        :rtype: ``SimpleEvent`` or ``Event``"""

        if len(outcomes) == 1:
            if callable(outcomes[0]):
                f = outcomes[0]
                outcomes = [
                 e.outcome for e in self._simple_events if f(e.outcome)
                ]
            else:
                for event in self._simple_events:
                    if event.outcome == outcomes[0]: return event
                else: return None
        simple = [e for e in self._simple_events if e.outcome in outcomes]
        return Event(*simple, name=name) if name else Event(*simple)


    def chances_of(self, *outcomes):
        """Returns the probability of the given outcome or outcomes occuring in
        a single statistical experiment.

        If multiple outcomes are given, the function will return the
        :py:class:`.Event` corresponding to one of those outcomes occuring.

        If a callable is provided, the function will return the
        :py:class:`.Event` of all the simple events that return ``True`` when
        the callable is applied to their outcome.

        :param \*outcomes: The outcome(s) to look for.
        :rtype: ``float``"""

        event = self.event(*outcomes)
        return event.probability() if event is not None else 0


    def experiment(self):
        """Generate an outcome."""

        outcomes = self.outcomes(p=True)
        outcomes = [(o, p) for o, p in outcomes.items()]
        outcomes, odds = zip(*outcomes)
        return random.choices(outcomes, weights=odds, k=1)[0]



"""Contains the Dataset class."""

from .variables import Variable

class Dataset:
    """A collection of :py:class:`.Variable` objects which describe the same
    experimental units.

    :param \*variables: The Variables that make up the Dataset.
    :raises TypeError: if non-Variables are given."""

    def __init__(self, *variables):
        for variable in variables:
            if not isinstance(variable, Variable):
                raise TypeError("{} is not a Variable".format(variable))
            if variable.length != variables[0].length:
                raise ValueError(
                 "Can't make Dataset with different-length Variables"
                )
        self._variables = list(variables)


    def __repr__(self):
        return "<Dataset ({} Variables)>".format(len(self._variables))


    @property
    def variables(self):
        """Returns the :py:class:`.Variable` objects in the Dataset.

        :rtype: ``tuple``"""

        return tuple(self._variables)


    def add_variable(self, variable):
        """Adds a :py:class:`.Variable` column to the Dataset.

        :param Variable variable: The Variable to add.
        :raises TypeError: if a non-Variable is given.
        :raises ValueError: if the Variable's length doesn't match."""

        if not isinstance(variable, Variable):
            raise TypeError("{} is not a Variable".format(variable))
        if self._variables and variable.length != self._variables[0].length:
            raise ValueError("Can't have Dataset of different-length Variables")
        self._variables.append(variable)


    def insert_variable(self, index, variable):
        """Inserts a :py:class:`.Variable` column to the Dataset at the index
        given..

        :param int index: The location to insert at.
        :param Variable variable: The Variable to add.
        :raises TypeError: if a non-Variable is given.
        :raises ValueError: if the Variable's length doesn't match."""

        if not isinstance(variable, Variable):
            raise TypeError("{} is not a Variable".format(variable))
        if self._variables and variable.length != self._variables[0].length:
            raise ValueError(
             "Can't have Dataset with different-length Variables"
            )
        self._variables.insert(index, variable)


    def remove_variable(self, variable):
        """Removes a :py:class:`.Variable` column from the Dataset.

        :param Variable variable: the Variable to remove."""

        self._variables.remove(variable)


    def pop_variable(self, index=-1):
        """Removes and returns the variable at a given index - by default the
        last one.

        :param int index: The index to remove at (default is ``-1``).
        :returns: the specified Variable."""

        return self._variables.pop(index)


    @property
    def rows(self):
        """Returns the rows of the Dataset.

        :rtype: ``tuple``"""

        columns = [var._values for var in self._variables]
        return tuple([values for values in zip(*columns)])


    def add_row(self, row):
        """Adds a row to the Dataset, updating all the Variables in the process.

        :param list row: A list of values to add.
        :raises ValueError: if the length of the row does not equal the number\
        of Variables in the Dataset."""

        if len(row) != len(self._variables):
            raise ValueError("Row {} is not the correct length".format(row))
        for value, variable in zip(row, self._variables):
            variable.add(value)


    def sort(self, column=None):
        """Sorts all the Variables in the Dataset by a single column, by
        default the first one.

        :param Variable column: the Variable to sort by.
        :raises TypeError: if a non-Variable is given.
        :raises ValueError: if the Variable given isn't in the Dataset."""

        var = self._variables[0]
        if column:
            if not isinstance(column, Variable):
                raise TypeError("Can't sort by non-Variable {}".format(column))
            if column not in self._variables:
                raise ValueError("{} isn't a Variable in {}".format(column, self))
            var = column
        indeces = list(range(len(var._values)))
        indeces.sort(key=var._values.__getitem__)
        for variable in self._variables:
            variable._values = list(map(variable._values.__getitem__, indeces))



"""Contains the base Variable class."""

from collections import Counter
from math import sqrt
from .exceptions import EmptyVariableError

class Variable:
    """A Variable represents an ordered sequence of measurements. It is `not`
    the same as a Python variable - it represents variables in the statistics
    sense of the word.

    A Variable is a container `and` an iterable of its values, and in many
    respects behaves like a ``list``.

    :param \*values: The values to go into the Variable. These will usually be\
    numerical, but can be any type. If you provide one value, which is iterable,\
    and which isn't a string, the values of that iterable will become the values\
    of the Variable.
    :param str name: The name of the Variable.
    :raises EmptyVariableError: if no values are given.
    :raises TypeError: if the name given isn't a string."""

    def __init__(self, *values, name=""):
        if len(values) == 0:
            raise EmptyVariableError("Cannot create Variable with no values")
        if len(values) == 1 and not isinstance(values[0], str):
            try:
                self._values = list(values[0])
                return
            except: pass
        self._values = list(values)
        if not isinstance(name, str):
            raise TypeError("name '{}' is not a str".format(name))
        self._name = name


    def __repr__(self):
        if self._name:
            return "<Variable '{}' {}>".format(self._name, tuple(self._values))
        return "<Variable {}>".format(tuple(self._values))


    def __len__(self):
        return len(self._values)


    def __contains__(self, member):
        return member in self._values


    def __iter__(self):
        return iter(self._values)


    def __getitem__(self, key):
        return self._values[key]


    def __setitem__(self, key, value):
        self._values[key] = value


    @property
    def values(self):
        """Returns the values in the Variable.

        :rtype: ``tuple``"""

        return tuple(self._values)


    def add(self, value):
        """Adds a value to the end of the Variable.

        :param value: The value to add."""

        self._values.append(value)


    def insert(self, index, value):
        """Inserts a value into the Variable.

        :param int index: The index to insert at.
        :param value: The value to insert."""

        self._values.insert(index, value)


    def remove(self, value):
        """Removes a value from the Variable.

        :param value: The value to remove.
        :raises EmptyVariableError: if you try to remove the only value."""

        if len(self._values) == 1:
            raise EmptyVariableError("Cannot remove last value from Variable")
        self._values.remove(value)


    def pop(self, index=-1):
        """Removes and returns the value at a given index - by default the last
        object in the Variable.

        :param int index: The index to remove at.
        :raises EmptyVariableError: if you try to pop the only value.
        :returns: the specified value."""

        if len(self._values) == 1:
            raise EmptyVariableError("Cannot pop last value from Variable")
        return self._values.pop(index)


    @property
    def name(self):
        """Returns the name of the Variable.

        :raises TypeError: if the name set is not a string."""

        return self._name


    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError("name '{}' is not a str".format(name))
        self._name = name


    @property
    def length(self):
        """The length of the Variable - the number of values it has.

        :rtype: ``int``"""

        return len(self)


    @property
    def max(self):
        """Returns the largest value."""

        return max(self._values)


    @property
    def min(self):
        """Returns the smallest value."""

        return min(self._values)


    @property
    def sum(self):
        """Returns the sum of the values."""

        return sum(self._values)


    @property
    def mean(self):
        """Returns the mean of the values - their sum divided by the number of
        values."""

        return self.sum / self.length


    @property
    def median(self):
        """Returns the median value - the value that occurs midway through
        when the values are sorted. If there is an even number, the midpoint
        between the two median values will be returned."""

        values = sorted(self._values)
        if len(values) % 2:
            return values[int((len(values) - 1) / 2)]
        midway = int(self.length / 2)
        return (values[midway - 1] + values[midway]) / 2


    @property
    def frequencies(self):
        """Returns the frequencies of the values in the Variable.

        :rtype: ``Counter``"""

        return Counter(self._values)


    @property
    def mode(self):
        """Returns the mode value - the value that occurs the most often. If
        more than one value meets this criteria, ``None`` is returned."""

        values = self.frequencies
        highest_frequency = max(values.values())
        if len([v for v in values if values[v] == highest_frequency]) == 1:
            return values.most_common()[0][0]


    @property
    def range(self, error=False):
        """Returns the range of the values - the difference between the
        largest and smallest values."""

        return self.max - self.min


    def variance(self, population=False):
        """Returns the variance of the values - the mean square deviation of
        the values from the mean. The values really have to be numerical for
        this to be meaningful.

        You can elect to get the population variance if you wish, which uses `N`
        rather than `N - 1` as the denominator.

        :param bool population: If ``True``, the population variance will be\
        returned (default is ``False``).
        :rtype: ``float``"""

        return sum([
         (value - self.mean) ** 2 for value in self._values
        ]) / (self.length - (not population))


    def st_dev(self, population=False):
        """Returns the standard deviation of the values, the square root of
        the :py:meth:`.variance` and a measure of deviation from the mean.
        As with that metric, you need numerical data for this to be sensible.

        You can elect to get the population deviation if you wish, which uses
        `N` rather than `N - 1` as the denominator.

        :param bool population: If ``True``, the population deviation will be\
        returned (default is ``False``).
        :rtype: ``float``"""

        return sqrt(self.variance(population=population))


    def zscore(self, value, population=False):
        """The z-score of a value is how many standard deviations it is from
        the mean.

        :param value: The value who's z-score you want to know.
        :param bool population: If ``True``, the population deviation will be\
        used (default is ``False``)."""

        return (value - self.mean) / self.st_dev(population=population)


    def covariance_with(self, variable):
        """Returns the covariance between this Variable and another Variable.
        This is a measure of how the variance of the two series reflect each
        other, and is a measure of correlation.

        :param Variable variable: The other Variable. It must be the same\
        length as this one.
        :raises TypeError: if something other than a Variable is given.
        :raises ValueError: if Variables of different length are given."""

        if not isinstance(variable, Variable):
            raise TypeError("{} is not a Variable".format(str(variable)))
        if self.length != variable.length:
            raise ValueError(
             "length {} is not length {}".format(self.length, variable.length)
            )
        this_mean, other_mean = self.mean, variable.mean
        square_deviations = sum([(value - this_mean) * (other - other_mean)
         for value, other in zip(self._values, variable._values)])
        mean_square_deviation = square_deviations / (self.length - 1)
        return mean_square_deviation


    def correlation_with(self, variable):
        """Returns the correlation of one Variable with another. This differs
        from :py:meth:`covariance_with` in that it is normalised to be
        between -1 and 1, so the maginitude of the result is important, rather
        than just the sign as is the case with covariance.

        All the same requirements apply - the object given must be a Variable,
        and they must be the same length.

        :param Variable variable: The other Variable. It must be the same\
        length as this one."""

        covariance = self.covariance_with(variable)
        sd_product = self.st_dev() * variable.st_dev()
        return covariance / sd_product



"""Contains tools for combining and permutating collections."""

from math import factorial
from functools import reduce
import operator
import itertools

def permutations(n, r=None):
    """Returns the number of ways of arranging r elements of a set of size n in
    a given order - the number of permuatations.

    :param int n: The size of the set containing the elements.
    :param int r: The number of elements to arange. If not given, it will be\
    assumed to be equal to n.
    :raises TypeError: if non-integers are given.
    :raises ValueError: if r is greater than n.
    :rtype: ``int``"""

    if not isinstance(n, int): raise TypeError("n {} must be integer".format(n))
    if r is None: return factorial(n)
    if not isinstance(r, int): raise TypeError("r {} must be integer".format(r))
    if r > n:
        raise ValueError("r {} is larger than n {}".format(r, n))
    return factorial(n) / factorial(n - r)


def combinations(n, r=None):
    """Returns the number of ways of combining r elements of a set of size n,
    where order doesn't matter

    :param int n: The size of the set containing the elements.
    :param int r: The number of elements to arange. If not given, it will be\
    assumed to be equal to n.
    :raises TypeError: if non-integers are given.
    :raises ValueError: if r is greater than n.
    :rtype: ``int``"""

    if not isinstance(n, int): raise TypeError("n {} must be integer".format(n))
    r = n if r is None else r
    if not isinstance(r, int): raise TypeError("r {} must be integer".format(r))
    if r > n:
        raise ValueError("r {} is larger than n {}".format(r, n))
    return factorial(n) / (factorial(r) * factorial(n - r))


def multiplications(*counts):
    """Returns the product of the integers given - useful for calculating the
    number of simple events in a multi-stage process.

    :raises TypeError: if non-integers are given.
    :rtype: ``int``"""

    if any(not isinstance(count, int) for count in counts):
        raise TypeError(f"Multiplication counts must be integers: {counts}")
    return reduce(operator.mul, counts, 1)


def permutate(collection, r=None):
    """Generates all the permutations of a given iterable, of a given length.

    It is essentially a wrapper around the built-in ``itertools.permutations``.

    :param iterable collection: The iterable to permutate.
    :param int r: The number of elements to arange. If not given, it will be\
    assumed to be equal to the length of the collection.
    :raises ValueError: if r is greater than the length of the collection.
    :rtype: ``tuple``"""

    n = len(tuple(collection))
    if r is not None and r > n:
        raise ValueError("r {} is larger than n {}".format(r, n))
    return itertools.permutations(collection, r=r)


def combine(collection, r=None):
    """Generates all the combinations of a given iterable, of a given length.

    It is essentially a wrapper around the built-in ``itertools.combinations``.

    :param iterable collection: The iterable to combine.
    :param int r: The number of elements to arange. If not given, it will be\
    assumed to be equal to the length of the collection.
    :raises ValueError: if r is greater than the length of the collection.
    :rtype: ``set``"""

    n = len(tuple(collection))
    if r is not None and r > n:
        raise ValueError("r {} is larger than n {}".format(r, n))
    r = n if r is None else r
    for combination in itertools.combinations(collection, r=r):
        yield frozenset(combination)


def multiply(*collections):
    """Generates all the multiplications of some iterables. For example, passing
    ``["A", "B"]`` and ``["+", "_"]`` would yield ``("A", "+")``,
    ``("A", "-")``, ``("B", "+")`` and ``("B", "-")``.

    :param \* collections: The iterables to put together.
    :rtype: ``tuple``"""

    for prod in itertools.product(*collections):
        yield prod



from setuptools import setup

setup(
 name="inferi",
 version="0.5.0",
 description="A statistics library.",
 url="https://inferi.samireland.com",
 author="Sam Ireland",
 author_email="mail@samireland.com",
 license="MIT",
 classifiers=[
  "Development Status :: 4 - Beta",
  "Intended Audience :: Science/Research",
  "License :: OSI Approved :: MIT License",
  "Topic :: Scientific/Engineering :: Mathematics",
  "Programming Language :: Python :: 3",
  "Programming Language :: Python :: 3.6",
 ],
 keywords="statistics probability data-science",
 packages=["inferi"],
 install_requires=["fuzz"]
)
